
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "myhtml.h"

enum {LIST_NONE=0, LIST_UL, LIST_OL, LIST_DL};

typedef struct hstate {
	int pre;	/* preformatted, default 0 = false */
	int list;	/* list mode, default LIST_NONE */
	int ll;		/* list level, starts with 1 */
	int indent;	/* indentation level, default 0 */
} hstate;

static int href;	/* links, default 0 */
static char **hrefs;

static void tag_a(hnode *, hstate *);
static void tag_address(hnode *, hstate *);
static void tag_area(hnode *, hstate *);
static void tag_b(hnode *, hstate *);
static void tag_body(hnode *, hstate *);
static void tag_br(hnode *, hstate *);
static void tag_dd(hnode *, hstate *);
static void tag_dl(hnode *, hstate *);
static void tag_dt(hnode *, hstate *);
static void tag_h1(hnode *, hstate *);
static void tag_h2(hnode *, hstate *);
static void tag_h3(hnode *, hstate *);
static void tag_h4(hnode *, hstate *);
static void tag_h5(hnode *, hstate *);
static void tag_h6(hnode *, hstate *);
static void tag_head(hnode *, hstate *);
static void tag_hr(hnode *, hstate *);
static void tag_html(hnode *, hstate *);
static void tag_i(hnode *, hstate *);
static void tag_img(hnode *, hstate *);
static void tag_li(hnode *, hstate *);
static void tag_map(hnode *, hstate *);
static void tag_ol(hnode *, hstate *);
static void tag_p(hnode *, hstate *);
static void tag_pre(hnode *, hstate *);
static void tag_table(hnode *, hstate *);
static void tag_td(hnode *, hstate *);
static void tag_th(hnode *, hstate *);
static void tag_tr(hnode *, hstate *);
static void tag_ul(hnode *, hstate *);
static void tag_unknown(hnode *, hstate *);
static void text(hnode *, hstate *);

static struct {
        char *name;
        void (*action)(hnode *, hstate *);
} tag[] = {
	{"A", tag_a},
	{"ADDRESS", tag_address},
	{"AREA", tag_area},
	{"B", tag_b},
	{"BR", tag_br},
	{"BODY", tag_body},
	{"DD", tag_dd},
	{"DL", tag_dl},
	{"DT", tag_dt},
	{"H1", tag_h1},
	{"H2", tag_h2},
	{"H3", tag_h3},
	{"H4", tag_h4},
	{"H5", tag_h5},
	{"H6", tag_h6},
	{"HEAD", tag_head},
	{"HR", tag_hr},
	{"HTML", tag_html},
	{"I", tag_i},
	{"IMG", tag_img},
	{"LI", tag_li},
	{"MAP", tag_map},
	{"OL", tag_ol},
	{"P", tag_p},
	{"PRE", tag_pre},
	{"TABLE", tag_table},
	{"TD", tag_td},
	{"TH", tag_th},
	{"TR", tag_tr},
	{"UL", tag_ul},
        {NULL, tag_unknown}
};

/* p looks like this: key=value key2="value 2". Keys are not case
   sensitive. Everything not in quotes is converted to upper case.
   Returns value if found, otherwise NULL.
*/
static char *get_value(char *k, char *p, char *b)
{
	char *q = p;
	int n = strlen(k);
	int i;

Again:
	while (*q && isspace(*q)) q++;	/* skip leading space */
	if (!strncasecmp(k, q, n) && q[n] == '=') {	/* gotcha */
		n++;
		i = 0;
		if (q[n] == '"') {
			n++;
			while (q[n] && q[n] != '"') b[i++] = q[n++];
		} else {
			while (q[n] && !isspace(q[n])) b[i++] = q[n++];
		}
		b[i] = '\0';
		return b;
	}
	while (*q && !isspace(*q)) q++;
	if (*q) goto Again;
	return NULL;
}

static int store_href(char *p)
{
	char b[1024];
	if (get_value("HREF", p, b)) {
		href++;
		hrefs = realloc(hrefs, (href+1) * sizeof *hrefs);
		hrefs[href] = strdup(b);
		return href;
	}
	return 0;
}

static hstate *new_hstate(void)
{
	hstate *s = malloc(sizeof *s);
	s->pre = 0;
	s->indent = 0;
	s->list = LIST_NONE;
	s->ll = 0;
	return s;
}

static void free_hstate(hstate *s)
{
	free(s);
}

static void newline(hstate *s)
{
	int i;

	putchar('\n');
	for (i = 0; i < s->indent; i++)
		putchar(' ');
}

static void tree(hnode *h, hstate *s)
{
	void (*action)(hnode *, hstate *);
	int i;

	if (h == NULL) return;
	if (h->tag == NULL) {
		text(h, s);
	} else {
		for (i = 0; tag[i].name; i++) {
			if (!strcmp(tag[i].name, h->tag)) break;
		}
		action = tag[i].action;
		(*action)(h, s);
	}
}

/* tag handlers are in alphabetical order so I can find them... */

/* Anchor tag. */
static void tag_a(hnode *h, hstate *s)
{
	if (store_href(h->text) != 0) printf("[%d]", href);
	tree(h->child, s);
	tree(h->next, s);
}

/* Address tag. Only changes font. */
static void tag_address(hnode *h, hstate *s)
{
	tree(h->child, s);
	tree(h->next, s);
}

/* Treat these as <a> */
static void tag_area(hnode *h, hstate *s)
{
	if (store_href(h->text) != 0) printf("[%d]", href);
	tree(h->child, s);
	tree(h->next, s);
}

static void tag_b(hnode *h, hstate *s)
{
	tree(h->child, s);
	tree(h->next, s);
}

/* Check for background values et al */
static void tag_body(hnode *h, hstate *s)
{
	tree(h->child, s);
	tree(h->next, s);
}

/* Line break */
static void tag_br(hnode *h, hstate *s)
{
	newline(s);
	tree(h->child, s);
	tree(h->next, s);
}

/* description in description list */
static void tag_dd(hnode *h, hstate *s)
{
	tree(h->child, s);		/* should be a bunch of text */
	tree(h->next, s);
}

/* start description list */
static void tag_dl(hnode *h, hstate *s)
{
	tree(h->child, s);
	tree(h->next, s);
}

/* tag in description list */
static void tag_dt(hnode *h, hstate *s)
{
	tree(h->child, s);
	tree(h->next, s);
}

/* all headers set font (unimplemented), start a new line, print
   their children and start a new line
*/
static void tag_h1(hnode *h, hstate *s)
{
	newline(s);
	tree(h->child, s);
	newline(s);
	tree(h->next, s);
}

static void tag_h2(hnode *h, hstate *s)
{
	newline(s);
	tree(h->child, s);
	newline(s);
	tree(h->next, s);
}

static void tag_h3(hnode *h, hstate *s)
{
	newline(s);
	tree(h->child, s);
	newline(s);
	tree(h->next, s);
}

static void tag_h4(hnode *h, hstate *s)
{
	newline(s);
	tree(h->child, s);
	newline(s);
	tree(h->next, s);
}

static void tag_h5(hnode *h, hstate *s)
{
	newline(s);
	tree(h->child, s);
	newline(s);
	tree(h->next, s);
}

static void tag_h6(hnode *h, hstate *s)
{
	newline(s);
	tree(h->child, s);
	newline(s);
	tree(h->next, s);
}

/* ignore this tag and all children */
static void tag_head(hnode *h, hstate *s)
{
	tree(h->next, s);
}

/* Horizontal ruler */
static void tag_hr(hnode *h, hstate *s)
{
	newline(s);
	printf("------------------------------------------------------------");
	newline(s);
	tree(h->child, s);
	tree(h->next, s);
}

/* this tag doesn't do anything per se */
static void tag_html(hnode *h, hstate *s)
{
	tree(h->child, s);
	tree(h->next, s);
}

static void tag_i(hnode *h, hstate *s)
{
	tree(h->child, s);
	tree(h->next, s);
}

/* Image tag. Ignored until we have somewhere to show pictures. */
static void tag_img(hnode *h, hstate *s)
{
	printf("[INLINE]");
	tree(h->child, s);
	tree(h->next, s);
}

/* start a new line. Should also fiddle with indentation */
static void tag_li(hnode *h, hstate *s)
{
	hstate ns = *s;
	ns.indent -= 2;
	newline(&ns);
	switch (ns.list) {
	case LIST_UL:
		printf("* ");
		break;
	case LIST_OL:
		printf("%d. ", ns.ll);
		break;
	default:
		printf("  ");
	}
	ns.indent += 2;
	ns.ll++;
	tree(h->child, &ns);
	tree(h->next, &ns);
}

/* Not much to do here... */
static void tag_map(hnode *h, hstate *s)
{
	tree(h->child, s);
	tree(h->next, s);
}

static void tag_ol(hnode *h, hstate *s)
{
	hstate ns = *s;
	ns.list = LIST_OL;
	ns.ll = 1;
	ns.indent += 4;
	newline(&ns);
	tree(h->child, &ns);
	newline(s);
	tree(h->next, s);
}

/* Paragraph break */
static void tag_p(hnode *h, hstate *s)
{
	newline(s);
	tree(h->child, s);
	tree(h->next, s);
}

/* Set a flag that this is preformatted. */
static void tag_pre(hnode *h, hstate *s)
{
	hstate ns = *s;
	ns.pre = 1;
	newline(&ns);
	tree(h->child, &ns);
	newline(s);
	tree(h->next, s);
}

/* We don't really do tables just yet. */
static void tag_table(hnode *h, hstate *s)
{
	tree(h->child, s);
	tree(h->next, s);
}

/* Table cell. One complication is that cells on the same row can be
   siblings or children in the parse tree, and that is correct HTML.
*/
static void tag_td(hnode *h, hstate *s)
{
	putchar('\t');
	tree(h->child, s);
	tree(h->next, s);
}

/* Table header */
static void tag_th(hnode *h, hstate *s)
{
	putchar('\t');
	tree(h->child, s);
	tree(h->next, s);
}

/* Table row */
static void tag_tr(hnode *h, hstate *s)
{
	newline(s);
	tree(h->child, s);
	tree(h->next, s);
}

/* To do anything useful with this, we need some state to tell us
   we're in a list and what kind the list is. IOW, add a struct
   hstate or something. Also add indentation level for children.
*/
static void tag_ul(hnode *h, hstate *s)
{
	hstate ns = *s;
	ns.list = LIST_UL;
	ns.ll = 1;
	ns.indent += 4;
	newline(&ns);
	tree(h->child, &ns);
	newline(s);
	tree(h->next, s);
}

/* Handles tags we don't have handlers for. Ignore the tag, do children */
static void tag_unknown(hnode *h, hstate *s)
{
	/* ignore the tag */
	tree(h->child, s);
	tree(h->next, s);
}

/* Copy text onto screen. Treats whitespace differently if this
   is "preformatted".
*/
static void text(hnode *h, hstate *s)
{
	int i;
	for (i = 0; h->text[i]; i++) {
		if (s->pre) putchar(h->text[i]);
		else if (isspace(h->text[i])) putchar(' ');
		else putchar(h->text[i]);
	}
	/* can't have children */
	tree(h->next, s);
}

static int nextc(void *closure)
{
	FILE *fp = closure;
	int c = getc(fp);
	if (c == EOF) return '\0';
	return c;
}

static void usage(void)
{
	printf("usage: lmb url\n");
	exit(0);
}

/* to contains the current url which is overwritten with the new one,
   which is calculated from to and from.

   normalize_url(<anything>, "http://localhost/") => "http://localhost/"
   normalize_url("http://localhost/index.html", "demo/index.html")
	=> "http://localhost/demo/index.html"
   normalize_url("http://localhost/
*/
void normalize_url(char *to, char *from)
{
	char pr1[10], ho1[1024], uri1[1024];
	char pr2[10], ho2[1024], uri2[1024];
	int po1, po2, m, n;

	m = parse_url(to, pr1, ho1, &po1, uri1);
	n = parse_url(from, pr2, ho2, &po2, uri2);
	if (n == 0) {
		/* everything there, ready to use */
		strcpy(to, from);
		return;
	} else if (from[0] == '/') {
		/* absolute uri */
		sprintf(to, "%s://%s:%d/%s", pr1, ho1, po1, from);
	} else {
		/* relative, cut to after first / */
		char *p = strrchr(to, '/');
		if (p) strcpy(p+1, from);
	}
}

int main(int argc, char **argv)
{
	hnode *h;
	hstate *s;
	int i, n, port;
	char proto[10], host[1024], uri[1024], url[1024], b[1024];
	char *file;
	FILE *fp;

	if (argc != 2) usage();
	strcpy(url, argv[1]);

	while (1) {
		printf("URL: %s\n\n", url);

		n = parse_url(url, proto, host, &port, uri);
		if (n != 0) usage();

		printf("proto='%s' host='%s' port=%d uri='%s'\n",
			proto, host, port, uri);

		if (!strcmp(proto, "http")) {
	                file = http_get(host, port, uri);
	        } else {
	                usage();
	        }
		href = 0;
		hrefs = NULL;

		fp = fopen(file, "r");
		if (fp == NULL) usage();

		h = parse_html(nextc, fp);
		if (h == NULL) usage();

		s = new_hstate();

		tree(h, s);
		newline(s);

		if (href) {
			newline(s);
			printf("References:\n\n");
			for (i = 1; i <= href; i++)
				printf("[%d] %s\n", i, hrefs[i]);
		}
		if (fgets(b, sizeof b, stdin) == NULL) return 0;

		if (isdigit(b[0]) && (n = atoi(b)) >= 1 && n <= href) {
			normalize_url(url, hrefs[n]);
		}

		free_html(h);
		free_hstate(s);
	}
	return 0;
}

